#version 330 core

#define MAXCOUNT 256

const float er = 1.0e10;

uniform bool centering;
uniform bool julia;
uniform int count;
uniform vec3 inflection[MAXCOUNT];

uniform vec2 center;
uniform vec2 radius;
uniform vec2 aspect;

uniform vec3 dragging;

in vec2 texcoord;

out vec4 colour;

vec2 cMul(vec2 a, vec2 b)
{
  return vec2(a.x * b.x - a.y * b.y, a.x * b.y + a.y * b.x);
}

vec2 cSqr(vec2 a)
{
  return cMul(a, a);
}

vec4 cMul(vec4 a, vec4 b)
{
  return vec4(cMul(a.xy, b.xy), cMul(a.xy, b.zw) + cMul(a.zw, b.xy));
}

vec4 cSqr(vec4 a)
{
  return cMul(a, a);
}

void main()
{
  vec2 texcoord_ = center + cMul(radius, aspect * vec2(texcoord));
  float px = float(length(vec4(radius, radius) * vec4(dFdx(vec2(aspect) * texcoord), dFdy(vec2(aspect) * texcoord))));
  vec4 c = vec4(texcoord_, px, 0.0);
  if (0 < count)
  {
    float r = inflection[0].z;
    vec4 f = vec4(inflection[0].xy, 0.0, 0.0);
    vec4 d = c;
    if (! centering)
    {
      d -= f;
    }
    c = cSqr(d / r) * r + f;
  }
  for (int i = 1; i < count && i < MAXCOUNT; ++i)
  {
    float r = inflection[i].z;
    vec4 f = vec4(inflection[i].xy, 0.0, 0.0);
    vec4 d = c - f;
    c = cSqr(d / r) * r + f;
  }
  int n = 0;
  vec4 z = julia && count != 0 ? c : vec4(0.0, 0.0, 0.0, 0.0);
        c = julia && count != 0 ? vec4(inflection[count - 1].xy, 0.0, 0.0) : c;
  for (n = 0; n < 1024; ++n)
  {
    if (dot(z.xy, z.xy) >= er)
      break;
    z = cSqr(z) + c;
  }
  if (dot(z.xy, z.xy) < er)
  {
    colour = vec4(1.0, 0.0, 0.0, 1.0);
    return;
  }
  float r = float(length(z.xy));
  float dr = float(length(z.zw));
  float de = 2.0 * r * log(r) / dr;
  float g = tanh(clamp(de, 0.0, 4.0));
  if (isnan(de) || isinf(de) || isnan(dr) || isinf(dr) || isnan(r) || isinf(r) || isnan(g) || isinf(g))
    g = 1.0;
  vec4 grey = vec4(vec3(g), 1.0);
  vec4 blue = vec4(0.5, 0.5, 1.0, 1.0);
  colour = length(texcoord_ - dragging.xy) < dragging.z ? blue * grey : grey;
}
